#include "gentrex.h"

/* special range values for loop fields */
enum {
	GTX_RANGE_INFINITE = -1
};

/* describe a range for a loop */
typedef struct gtx_range_s {
	int from;       /* between 0 and GTX_RANGE_INFINITE */
	int to;         /* between 0 and GTX_RANGE_INFINITE */
	int state_idx;  /* index in gentrex_state_t->search_state[] */
} gtx_range_t;

/* bytecode instruction opcodes */
typedef enum gtx_inst_type_e {
	INST_NODE_REF,           /* a reference to a node by name (read from the map) */
	INST_NOT,                /* '!' operator: negate the bool value of NUT */
	INST_TREE_CHILD,         /* '.' operator: first child of NUT (might be looped) */
	INST_TREE_CHILDREN,      /* '/' operator: any child of NUT (might be looped: BFS) */
	INST_TREE_PARENT,        /* '^' operator: first parent of NUT (might be looped) */
	INST_TREE_NEXT,          /* '>' operator: right sibling of NUT (might be looped) */
	INST_BRPUSH,             /* mid-condition brnahc for '&&' and '||'; after the left side has been evaluated, check if the right side can be skipped */
	INST_AND,                /* '&&' operator after both sides pushed onto the stack */
	INST_OR,                 /* '||' operator after both sides pushed onto the stack */
	INST_ASG,                /* assignment: copy NUT pointer to the map under some name */
	INST_TAG_EQ,             /* check whether a NUT tag equals to a string */
	INST_TAG_MATCH,          /* check whether a NUT tag matches a string */
	INST_TAG_C_OP,           /* check whether a NUT tag against a custom operator */
	INST_TAG_TRUE,           /* the '[]' expression: matches any existing NUT; it's a no-op */
	INST_ENTER,              /* save states and enter a parenthesis (on '(') */
	INST_LEAVE               /* restore states and leave a parenthesis (on ')') */
} gtx_inst_type_t;

/* byte code instructions with optional operands */
typedef struct gtx_inst_s gtx_inst_t;
struct gtx_inst_s {
	gtx_inst_type_t inst;    /* opcode */

	union {                  /* operand 1 */
		char *tagn;
		gtx_inst_type_t brtype;
		gtx_range_t range;
	} arg1;

	union {                  /* operand 2 */
		char *str;
		gtx_inst_t *branch;
	} arg2;

	void *match_cookie;
	char *c_op;             /* custom operator */
};

/* bytecode: a sequence of instructions in an array and metadata to track
   length of the array */
typedef struct gtx_code_s {
	int alloced, used;
	gtx_inst_t *code;
} gtx_code_t;

/* entry of an open list: used when a search needs to keep a list of
   nodes to be visited later. E.g. a '/' operator does a BFS and needs
   to keep track on the next level to be visited. */
typedef struct gtx_ol_s {
	gentrex_nodecomp_t ndc;
	int depth;
} gtx_ol_t;

/* each looped operator has a search state that can be incremented. The
   iteration goes on until the expression has been evaluated to all
   valid search state combinations */
typedef struct gtx_search_state_s {
	int depth;                /* matched against range to decide if the result is of interest in a loop; the exact interpretation depends on the operator */
	gentrex_nodecomp_t ndc;   /* current NUT for this search */
	int inc;                  /* whether an increment should happen during the evaluation */

	/* open list of nodes-to-be-visited */
	int ol_used, ol_alloced;  /* array sizes */
	int ol_head;              /* next element from the front that hasn't been removed */
	gtx_ol_t *ol;             /* the actual array of elements */
} gtx_search_state_t;

/* expression evaluation state; this holds all states required for going on
   with the next iteration of evaluating the expression in subsequent calls. */
struct gentrex_state_s {
	const char *err_msg;         /* error message from flex/bison */
	void *scanner;               /* flex state */
	gtx_code_t *code;            /* bytecode of the expression */
	char *match_c_op;            /* custom operator string for the match operator (sep char) */
	char *sval;                  /* pointer to yytext of an integer token */

	const gentrex_api_t *api;    /* glue to the caller to handle the tree and the map */

	/* search state: size known after compiling the bytecode */
	int search_state_len;
	gtx_search_state_t *search_state;

	int reached_end; /* wether the search has already reached its end - if true, subsequent increment/execute won't yield new result */

	/* eval stack: always reset by the evaluator, it's still worth keeping it
	   accross iterations: it's likely that a subsequent iteration won't take
	   much more stack than the previous one: allocate once, use until the
	   state is deleted. */
	int sp, sa;  /* stack pointer, stack allocation */
	gentrex_nodecomp_t *stack;
};

/* allocate and append instruction to the current bytecode; operands
   are not set; where required, they should be overwritten by the
   caller after the call. */
gtx_inst_t *gtx_append(gentrex_state_t *ctx, gtx_inst_type_t it);
